#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <errno.h>
#include <unistd.h>
#include <netdb.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <inttypes.h>

#include "idip.h"
#include "message.h"
#include "trace.h"

/*
 * Sarah Jelinek
 * University of Colorado, Colorado Springs
 * Computer Science Masters Project.
 * 5/17/03
 * Copyright 2003
 */

#define DEBUG_DEMO 1 
#define DEBUG 1
#define MODNAME "idip_firewall_receiver.c"
#define LINESIZE	120
static int process_idip_message(idip_message_t *msg);
static int do_trace_request(idip_app_msg_t *msg);
static int do_request(idip_app_msg_t *msg);
static int undo_request(idip_app_msg_t *msg);
/*
 * The backplane listens on a socket and determines the type of request
 * being sent to it. From there it invokes the appropriate processing.
 */

void
main() {

	int			length;
	int			n;
	idip_message_t		i_message;
	struct sockaddr_in	toaddr;

	/* Set up our listening socket */
	if ((gen_mbx = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
		fprintf(stderr, "Unable to set up receiver socket.\n");
		perror(strerror(errno));
		return;
	}
	/*
	 * Listen for messages from any host, on the IDIP_APP_PORT
	 */
	(void) memset(&gen_from, 0, sizeof (gen_from));
	gen_from.sin_family = AF_INET;
	gen_from.sin_addr.s_addr = INADDR_ANY;
	gen_from.sin_port = htons(IDIP_APP_PORT);

	if (bind(gen_mbx, (struct sockaddr *) &gen_from,
		sizeof (struct sockaddr_in)) < 0) {
		fprintf(stderr, "%s", "Could not bind to port\n");
		perror(strerror(errno));
	}

	length = sizeof (gen_from);

	if (getsockname(gen_mbx, (struct sockaddr *) &gen_from, &length)) {
		perror("getting socket name");
		exit(1);
	}

	while (1) {
#ifdef DEBUG_DEMO
printf("%s: Waiting for incoming idip messages on firewall\n", MODNAME);
#endif
		n = recvfrom(gen_mbx, &i_message, 
		    sizeof (idip_message_t), 
		    0, (struct sockaddr *)&gen_from, &length);
#ifdef DEBUG
printf("idip_message_receiver: sizeof idip_message_t %d\n", 
    sizeof (idip_message_t));
printf("idip_message_receiver: n bytes received :%d\n", n);
printf("idip_message_receiver: message received: %s\n", 
    i_message.p.app_msg.data);
#endif
		
		if (n < 0) {
		    perror("receiving datagram messages");
		    continue;
		}
#ifdef DEBUG
printf("idip_message_receiver: received idip_message packet\n");
#endif
		/*
		 * Process this message. It is possible that there has
		 * been a transmission problem or data is garbled. Move on
		 * if this is the case.
		 */
		if (process_idip_message(&i_message) != 0) {
			perror("error processing idip message");
			continue;
		}
	}

	/*NOTREACHED*/
}

static int
process_idip_message(idip_message_t *msg)
{
	int error = 0;
	idip_app_msg_t app_msg = msg->p.app_msg;

	if (msg->i_hdr.next_type == IDIP_APPLICATION_DATA) {
		if (msg->p.app_msg.app_hdr.class_id == IDIP_APP_CLASS_TRACE) {
			error = do_trace_request(&(msg->p.app_msg));
			if (error) {
			    perror("executing trace request");
			    return (-1);
			}
		} else if (msg->p.app_msg.app_hdr.class_id == 
			IDIP_APP_CLASS_DCDO) {
#ifdef DEBUG_DEMO
fprintf(stderr, "%s process_idip_message: Received DCDO request\n", MODNAME);
fflush(stderr);
#endif
			error = do_request(&(msg->p.app_msg));
			if (error) {
			    perror("executing do request");	
			    return (-1);
			}
		} else if (msg->p.app_msg.app_hdr.class_id == 
			IDIP_APP_CLASS_DCUNDO) {
#ifdef DEBUG_DEMO
fprintf(stderr, "%s process_idip_message: Received DCUNDO request\n", MODNAME);
fflush(stderr);
#endif
			error = undo_request(&(msg->p.app_msg));
			if (error) {
			    perror("executing undo request");
			    return (-1);
			}
		}
	} else {
		printf("Unrecognized app header type\n");
		return (-1);
	}
	return (0);
		    
}
/*
 * Start the tracing, via tcpdump. Determine if we are under attack and if
 * so notify upstream router, and ourselves to block traffic from the 
 * sources we  found.
 */
static int
do_trace_request(idip_app_msg_t *msg)
{
	FILE    *fp;
        FILE    *fp1;
	int	snd;
        char    buf[BUFSIZ];
        char    buf1[BUFSIZ];
        char    *cmd; 
        char    *cmd1;
	int	duration = 10;
	int	status;
	FILE	*netfp;
	char	interface[80];
	char	tinterface[80];
	char	router[80];
	char	source[80];
	char	line[BUFSIZ];
	int	sent = 0;
	int	number;
	int	count = 0;
	int	len, len1;
	int	ret = 0;
	int	threshold;
	char	*t;
	char	*tok = NULL;
	char    alertmsg[200]; 

	idip_message_t idip_do_msg;
	struct sockaddr_in snd_addr;
	struct 	in_addr haddr;
	char	*thishost;
	struct  hostent *hent = NULL;

	hent =  gethostent(); /* this host is the one being attacked */

	if (hent != NULL)
		thishost = hent->h_name;
#ifdef DEBUG_DEMO
printf("%s do_trace_request: thishost is %s\n", MODNAME, thishost);
#endif

	memset(&snd_addr, 0, sizeof (snd_addr));
	snd_addr.sin_family = AF_INET;
	/*
	 * On the upstream router, idip_firewall_receiver is listening on
	 * this port.
	 */
	snd_addr.sin_port = htons(IDIP_APP_PORT);

	/*
	 * setup snd socket. idip_firewall_receiver listens on a dgram
	 * type of socket. 
	 */
	if ((snd = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
		perror("socket");
		return (-1);
	}

	/*
	 * Read the message the max number of connections allowed in that time.
	 * THRESHOLD is the keyword here.
	 */
#ifdef DEBUG_DEMO
printf("received message %s\n", msg->data);
#endif
	t = strstr(msg->data, "THRESHOLD");
	tok = strtok(t, " ");
	tok = strtok(NULL, " ");
	if (tok == NULL)
		tok = "50";

#ifdef DEBUG_DEMO
printf("threshold packets %s\n", tok);
#endif
		
	cmd = malloc(1024);
	memset(buf, 0, sizeof (buf));
	memset(buf1, 0, sizeof (buf1));
	/*
	 * Packet limit is arbitrary, default is 150. Can be adjusted.
	 */
#ifdef DEBUG_DEMO
	duration = 10;
	sprintf(cmd, "sh tcpdump.sh %d %d", atoi(tok), 3);
	printf("cmd = %s\n", cmd);
#endif
	/*
	 * Run tcpdump on all interfaces for specified duration and
	 * threshold.
	 */

        if ((fp = popen(cmd, "r")) != NULL) {
                while (fgets(buf, BUFSIZ, fp) != NULL) {
			memset(interface, 0, sizeof (interface));
			memset(tinterface, 0, sizeof (tinterface));
			memset(router, 0, sizeof (router));
			memset(source, 0, sizeof (source));

			if (sscanf(buf, "%s %s %d", source, interface, 
			    &number)) {
				if (strcmp(source, "0.0.0.0") == 0) {
					continue;
				}
				 
#ifdef DEBUG_DEMO
printf("%s do_trace_request: UNDER ATTACK:\n", MODNAME);
printf("%s do_trace_request: from source %s\n", MODNAME, source);
printf("%s do_trace_request: on interface %s\n", MODNAME, interface);
printf("%s do_trace_request: number of packets %d\n", MODNAME, number);
#endif
			
			}
			if ((netfp = fopen("topo.txt", "r")) == NULL) {
				fprintf(stderr, "Cannot open topology file\n");
				exit (1);
			}
#ifdef DEBUG
printf("%s do_trace_request: Reading topo.txt file\n", MODNAME);
#endif
			while (fgets(line, LINESIZE, netfp) != NULL) {
#ifdef DEBUG
printf("%s do_trace_request: Line from topo.txt file read %s\n", MODNAME, 
   line);
#endif
				/*
				 * First send to the identified interface
				 * under attack.
				 */
				if (strstr(line, interface)) {
#ifdef DEBUG
printf("%s do_trace_request: Found interface in topo file\n", MODNAME);
#endif
					/*
					 * Get router info from conf file.
					 */
					if (sscanf(line, "%s %s", tinterface,
					    router) ) {
#ifdef DEBUG
printf("%s do_trace_request: interface %s has router %s\n", MODNAME, 
interface, router);
#endif
					} else {
						/*
						 * This shouldn't happen. All
						 * interfaces specified in
						 * topo file should have
						 * router specified.
						 */
						continue;
					}
				} else {
					/*
					 * Not interface we are looking for.
					 */
					continue;
				}
				/*
				 * Copy router inet_addr to the snd_addr 
				 * so routed appropriately.
				 */
				if ((inet_aton(router, &haddr)) == 0)
					fprintf(stderr, "haddr conversion "
					    " failed\n");

				bcopy(&haddr, &snd_addr.sin_addr, 
				    sizeof (haddr));

				(void) sprintf(alertmsg,
	"FLOOD DETECTED on %s from %s (THRESHOLD %d connections exceeded in %d seconds)", thishost, source, atoi(tok), duration);
				printf("alertmsg after sprintf %s\n", alertmsg);
				/*
				 * Create idip message to send to upstream
				 * router.
				 */
				if (create_do_msg(&idip_do_msg, alertmsg, 
				    thishost, source, IDIP_LIMIT_RATE) != 0) {
					fprintf(stderr, "Failed "
					    "to create idip request\n");
					continue;
				}
				/*
				 * Now send this to the router we found
				 * in the topology file.
				 */
				if (sendto(snd, (idip_message_t *)&idip_do_msg, 
				    sizeof (idip_message_t), 0, 
				    (struct sockaddr *)&snd_addr, 
				    sizeof (snd_addr)) < 0) {
					perror("no message sent");
				} 
				/*
				 * Now process this to ourselves. Send to
				 * Rate limiter on this host.
				 */
				if (do_request(&(idip_do_msg.p.app_msg)) != 0) {
					fprintf(stderr, "Could not notify " 
					    "current host rate limiter \n");
				}
#ifdef DEBUG_DEMO
printf("%s do_trace_request: alertmsg sent to %s: %s\n", MODNAME, router, alertmsg);
#endif
			} /* end 1st while  */
			fclose(netfp);
			netfp = NULL;
notify_others:

#ifdef DEBUG_DEMO
printf("%s do_trace_request : Checking for other upstream routers to notify\n",
    MODNAME);
#endif
			/*
			 * Reread the file and send to other interfaces
			 * that have not been notified.
			 */
			if ((netfp = fopen("topo.txt", "r")) == NULL) {
				fprintf(stderr, "Cannot open topology file\n");
				exit (1);
			}
			
			while (fgets(line, LINESIZE, netfp) != NULL) {
				if (strstr(line, interface)) {
					/*
					 * Already sent to this interface.
					 * Looking for others.
					 */
					continue;
				}
				if (sscanf(line, "%s %s", tinterface, 
				    router)) {
#ifdef DEBUG
printf("%s do_trace_request: line read from topo file %s\n", MODNAME, line);
#endif
					/* 
					 * craft alert and send
					 * it to the other interfaces
					 */
					if (inet_aton(router, 
					    (struct in_addr *)&haddr) == 0)
						perror("inet_aton");

					bcopy(&haddr, &snd_addr.sin_addr, 
					    sizeof (haddr));
	
					(void) sprintf(alertmsg,
	"FLOOD DETECTED on %s from %s (THRESHOLD %d connections exceeded in %d seconds)", thishost, source, atoi(tok), duration);
					/*
					 * Create idip message to send to
					 * upstream router.
					 */
					if (create_do_msg(&idip_do_msg, 
					    alertmsg, thishost, source, 
					    IDIP_LIMIT_RATE) != 0) {
						fprintf(stderr, "Failed "
						    "to create idip request\n");
						continue;
					}
					/*
					 * Now send this to the router we found
					 * in the topology file.
					 */
					if (sendto(snd, 
					    (idip_message_t *)&idip_do_msg, 
					    sizeof (idip_message_t), 0, 
					    (struct sockaddr *)&snd_addr, 
					    sizeof (snd_addr)) < 0) {
						perror("no message sent");
					}
#ifdef DEBUG_DEMO
printf("%s do_trace_request(): alertmsg sent to %s: %s\n", MODNAME, router, alertmsg);
#endif
				} /* end if sscanf */
				fclose(netfp);
			} /* End 2nd while */
		} /* end outer while */
	} /* End popen */
	(void) pclose(fp); /* Close pipe */
	close(snd);
	return (0);
}
/*
 * This actually handles the request to limit or restore the rate for specific
 * source.
 */
static int 
do_request(idip_app_msg_t *msg)
{
	struct	sockaddr_in snd_addr;
	char	*app_data;
	int   	snd;
	int	ret;
	in_addr_t haddr;

	app_data = msg->data;
#ifdef DEBUG_DEMO
fprintf(stderr, "%s do_request: message received %s\n", MODNAME, app_data);
fflush(stderr);
#endif
	/*
	 * Now send this message to the rate limiter on port 6779
	 * on this host.
	 */
	haddr = inet_addr("192.168.16.102");
	snd = socket(AF_INET, SOCK_STREAM, 0);
	if (snd < 0) {
		perror("opening data socket");
		return (-1);
	}

	bzero((char *)&snd_addr, sizeof (snd_addr));
	snd_addr.sin_family = AF_INET;
	snd_addr.sin_port = htons(IDIP_RTLIMIT_PORT);
	bcopy((char *)&haddr, (char *)&snd_addr.sin_addr,
	    sizeof (haddr));

	ret = connect(snd, (struct sockaddr *)&snd_addr, sizeof (snd_addr));
	if (ret == -1) {

#ifdef DEBUG_DEMO
printf("%s do_request: Could not connect to rate limiter\n", MODNAME);
#endif
		perror(strerror(errno));
		return (ret);
	} else {

#ifdef DEBUG_DEMO
printf("%s do_request: Connected to rate limiter\n", MODNAME);
#endif
	}
	if ((send(snd, (char *)app_data, strlen(app_data), 0)) < 0) {
#ifdef DEBUG_DEMO
printf("%s do_request: Could not send rate limiter request: %s\n", MODNAME,
    app_data);
#endif
		return (-1);
	}
#ifdef DEBUG_DEMO
printf("%s do_request: Sent msg %s to rate limiter\n", MODNAME, app_data);
#endif
	return (0);
}

static int 
undo_request(idip_app_msg_t *msg)
{
	struct	sockaddr_in snd_addr;
	char	*app_data;
	int   	snd;
	int	ret;
	char	interface[80];
	char	router[80];
	idip_message_t idip_do_msg;
	in_addr_t haddr;
	FILE	*netfp;
	char	line[BUFSIZ];
	char	*thishost;
	struct  hostent *hent = NULL;

	hent =  gethostent(); /* this host is the one being attacked */

	if (hent != NULL)
		thishost = hent->h_name;

#ifdef DEBUG_DEMO
fprintf(stderr, "%s undo_request(): Processing IDIP undo request\n", MODNAME);
fflush(stderr);
#endif

	/*
	 * Some basic sanity tests.
	 */
	if (msg == NULL) {
		perror("Message format error");
		return (-1);
	}

	/*
	 * This is a request from the IDS system to limit traffic on this
	 * host.
	 */ 
	app_data = msg->data;
#ifdef DEBUG_DEMO
fprintf(stderr, "%s undo_request: message received %s\n", MODNAME, app_data);
fflush(stderr);
#endif
	snd = socket(AF_INET, SOCK_DGRAM, 0);
	if (snd < 0) {
		perror("opening data socket");
		return (-1);
	}
	/*
	 * This is an end of flood request so we must send it to
	 * the upstream routers for processing. If no routers 
	 * configured, then break out of this loop.
	 */
	if ((netfp = fopen("topo.txt", "r")) != NULL) {
#ifdef DEBUG_DEMO
fprintf(stderr, "%s undo_request: opened topo file\n", MODNAME);
fflush(stderr);
#endif
		while (fgets(line, LINESIZE, netfp) != NULL) {
			memset(interface, 0, sizeof (interface));
			memset(router, 0, sizeof (router));
			/*
			 * Get router info from conf file.
			 */
			if (sscanf(line, "%s %s", interface, router)) {
#ifdef DEBUG_DEMO
fprintf(stderr, "%s undo_request: router found %s\n", MODNAME, router);
fflush(stderr);
#endif
				/*
				 * Copy router inet_addr to the snd_addr 
				 * so routed appropriately.
				 */
				if ((inet_aton(router, 
				    (struct in_addr *)&haddr)) == 0) {
					fprintf(stderr, "haddr conversion "
					    " failed\n");
					fflush(stderr);
					break;
				}
				bzero((char *)&snd_addr, sizeof (snd_addr));
				bcopy(&haddr, &snd_addr.sin_addr, 
				    sizeof (haddr));
				/*
				 * Create idip message to send to upstream
				 * router, if it is listening on the idip port.
				 */
				if (create_do_msg(&idip_do_msg, app_data, 
				    thishost, thishost, 
				    IDIP_RESTORE_RATE) != 0) {
					fprintf(stderr, "Failed "
					    "to create idip request\n");
					continue;
				}
				/*
				 * Now send this to the router we found
				 * in the topology file.
				 */

				snd_addr.sin_family = AF_INET;
				snd_addr.sin_port = htons(IDIP_APP_PORT);
				if (sendto(snd, (idip_message_t *)&idip_do_msg, 
				    sizeof (idip_message_t), 0, 
				    (struct sockaddr *)&snd_addr, 
				    sizeof (snd_addr)) < 0) {
					perror("no message sent");
				} 
#ifdef DEBUG_DEMO
fprintf(stderr, "%s undo_request: msg sent to %s: %s\n", MODNAME, router, app_data);
fflush(stderr);
#endif
			} 
		} /* end while */
	} /*end if End */
	close (snd);

	/*
	 * Now send this message to the rate limiter on port 6779
	 * on this host.
	 */
	snd = socket(AF_INET, SOCK_STREAM, 0);
	if (snd < 0) {
		perror("opening data socket");
		return (-1);
	}

	haddr = inet_addr("192.168.16.102");

	bzero((char *)&snd_addr, sizeof (snd_addr));
	snd_addr.sin_family = AF_INET;
	snd_addr.sin_port = htons(IDIP_RTLIMIT_PORT);
	bcopy((char *)&haddr, (char *)&snd_addr.sin_addr,
	    sizeof (haddr));

	ret = connect(snd, (struct sockaddr *)&snd_addr, sizeof (snd_addr));
	if (ret == -1) {

#ifdef DEBUG_DEMO
fprintf(stderr, "%s undo_request: Could not connect to rate limiter\n", MODNAME);
fflush(stderr);
#endif
		perror(strerror(errno));
		return (ret);
	} else {

#ifdef DEBUG_DEMO
fprintf(stderr, "%s undo_request: Connected to rate limiter\n", MODNAME);
fflush(stderr);
#endif
	}
	if ((send(snd, (char *)app_data, strlen(app_data), 0)) < 0) {

#ifdef DEBUG_DEMO
fprintf(stderr,"%s undo_request: Could not send rate limiter request: %s\n", MODNAME, app_data);
fflush(stderr);
#endif
		return (-1);
	}
#ifdef DEBUG_DEMO
fprintf(stderr, "%s undo_request: Sent msg %s to rate limiter\n", MODNAME, app_data);
fflush(stderr);
#endif
	close (snd);
	return (0);
}
