#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <netdb.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <errno.h>

#include "idip.h"
/*
 * Sarah Jelinek
 * University of Colorado, Colorado Springs
 * Computer Science Masters Project.
 * 5/17/03
 * Copyright 2003
 */

/*
 * Adds the address that is passed in to the ack list. If already there
 * increments the number of outstanding acks waiting.
 */
int 
add_ack_list(struct ack_list **ackl, uint32_t node)
{
	struct ack_list	*acp;
	struct ack_data *ackp;

	if (*ackl == NULL) {
	    ackl = malloc(sizeof (struct ack_list));
	    if (*ackl == NULL) {
		perror("mallocing memory for ack list");
		return (-1);
	    }
	    /*
	     * If the list was null, no entries, add this one.
	     */
	    ackp = (struct ack_data *)malloc(sizeof (struct ack_data));
	    ackp->node = node;
	    ackp->num_outstanding = 1;
	    (*ackl)->ah = *ackp;
	    (*ackl)->next = NULL;
	    return (0);
	}
	/*
	 * Search the current list to see if an entry exists. If not,
	 * allocate the memory and add the new data. Otherwise increase
	 * the num_outstanding value.
	 */
	acp = *ackl;
	while (acp != NULL) {
	    if (acp->ah.node == node) {
		acp->ah.num_outstanding++;
		return (0);
	    }
	    acp = acp->next;
	}
	/*
	 * If we are here, we did not find it, add it to the list.
	 */
	 
	ackp = (struct ack_data *)malloc(sizeof (struct ack_data)); 
	if (ackp == NULL) { 
		perror("mallocing for ack data struct");
	        return (-1);
	}
	ackp->node = node;
	ackp->num_outstanding = 1;
	acp->ah = *ackp;
	acp->next = NULL;
	return (0);

}

int 
remove_ack_list(struct ack_list **ackl, uint32_t node)
{
	
	struct ack_list	*acp;

	acp = *ackl;
	while (acp != NULL) {
	    if (acp->ah.node == node) {
		acp->ah.num_outstanding--;
		return (0);
	    }
	    acp = acp->next;
	}
	/*
	 * Otherwise not in this list and that is a problem.
	 */
	return (-1);
}
