#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <netdb.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <errno.h>

#include "idip.h"
#include "idip_message.h"
/*
 * Sarah Jelinek
 * University of Colorado, Colorado Springs
 * Computer Science Masters Project.
 * 5/17/03
 * Copyright 2003
 */

static int dc_list_alloc(struct dc_list **dcl);

/*
 * Looks for the discovery coordinator in the dc list. If found, returns 0.
 * If not found returns 1. Otherwise, a non-zero value is returned in error.
 */
int 
in_dc_list(struct dc_list *dcl, struct idip_hello_entry entry)
{
	int	error = 0;
	struct dc_list *dclp;

	if (dcl == NULL) {
		error = dc_list_alloc(&dcl);
		if (error) {
			return (error);
		}
		/* 
		 * If we had to allocate the list it isn't in there so return
		 * an error to indicate this.
		 */
		return (1);
	} 

	/*
	 * Otherwise search for it in the list
	 */
	dclp = dcl;
	
	while (dclp != NULL) {
		if (dclp->dc.ip_addr == entry.id) {
			return (1);
		}
		dclp = dclp->next;
	}
	return (error);

}
/*
 * Adds the address that is passed in to the dc list.
 */
int 
add_dc_list(struct dc_list **dcl, struct idip_hello_entry entry)
{

	struct dc_list *tmp = *dcl;
	struct dc_list *prev;

	struct dc_data *dcp = (struct dc_data *)malloc(sizeof (struct dc_data));

	if (dcp == NULL)
		return ENOMEM;

	/* Should I do a memcpy here? Probably */
	dcp->ip_addr = entry.id;
	prev = NULL;

	while (tmp) {
		prev = tmp;
		tmp = tmp->next;
	}

	if (prev == NULL) {
		tmp->next->dc = *dcp;
	} else {
		prev->next->dc = *dcp;
	}
		 
	return 0;
}

static int
dc_list_alloc(struct dc_list **dcl)
{
	*dcl = (struct dc_list *)malloc(sizeof (struct dc_list));

	if (*dcl == NULL) {
		return ENOMEM;
	}

	(void) memset(*dcl, 0, sizeof (struct dc_list));
	return (0);
}
