#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <netdb.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <errno.h>

#include "message.h"
#define DEBUG 1
/*
 * Sarah Jelinek
 * University of Colorado, Colorado Springs
 * Computer Science Masters Project.
 * 3/17/04
 * Copyright 2004
 */

static int get_address(char *message, char **address);

/*
 * For a trace message, the request is not set, but we pass it in
 * in case the trace message semantics change at any point.
 * Tracing is just one action for now.
 */
int
create_trace_msg(idip_message_t *msg, char *tracemsg, char *source, 
    char *this_host, int request)
{
	uint32_t host;

	host = inet_addr(this_host);

	msg->i_hdr.next_type = IDIP_APPLICATION_DATA;
	msg->i_hdr.dest_addr = 0;
	msg->i_hdr.length = sizeof(struct idip_header);
	msg->i_opt_hdr.next_type =  0;

#ifdef DEBUG
printf("message.c: create_trace_msg(): init\n");
#endif

	create_app_msg(&(msg->p.app_msg), IDIP_APP_CLASS_TRACE, host, source,
	    tracemsg, request);

#ifdef DEBUG
printf("create_trace_msg: tracemsg %s\n", msg->p.app_msg.data);
#endif
	return (0);
}

void
create_app_msg(idip_app_msg_t *msg, int type, uint32_t this_host, 
    char *source, char *appmsg, int request) 
{
	msg->app_hdr.class_id = type;
	msg->app_hdr.length = sizeof (*msg);
	msg->app_hdr.orig_addr.addr = this_host;

	int level = 0;

	sprintf(msg->data, appmsg);

#ifdef DEBUG
printf("In create_app_msg, type %d\n", type);
#endif
   
	msg->body.trace.data.cidd.ip_protocol =  0;
	msg->body.trace.data.cidd.ip_source = inet_addr(source);

	/*
	 * The do message looks at what type of request it is. 
	 */
	if (type == IDIP_APP_CLASS_DCUNDO) {
			msg->body.trace.req_action_hdr.limit_from = 0;
			msg->body.trace.req_action_hdr.limit_until = 0;
			msg->body.trace.req_action_hdr.burst_rate = 0;
			msg->body.trace.req_action_hdr.limit_action = 
			    IDIP_DROP;
			return;
	}

	if (type == IDIP_APP_CLASS_DCDO) {
		if (request == IDIP_LIMIT_RATE && level == 4) {
			msg->body.trace.req_action_hdr.limit_from = 0;
			msg->body.trace.req_action_hdr.limit_until = 10;
			msg->body.trace.req_action_hdr.burst_rate = 0;
			msg->body.trace.req_action_hdr.limit_action = 
			    IDIP_DROP;
		} else if (request == IDIP_LIMIT_RATE && level == 3) {
			msg->body.trace.req_action_hdr.limit_from = 0;
			msg->body.trace.req_action_hdr.limit_until = 30;
			msg->body.trace.req_action_hdr.burst_rate = 150;
			msg->body.trace.req_action_hdr.limit_action = 
			    IDIP_DROP; 
		} else if (request == IDIP_LIMIT_RATE && level == 2) {
			msg->body.trace.req_action_hdr.limit_from = 0;
			msg->body.trace.req_action_hdr.limit_until = 30;
			msg->body.trace.req_action_hdr.burst_rate = 50;
			msg->body.trace.req_action_hdr.limit_action = 
			    IDIP_DROP;
		} else if (request == IDIP_LIMIT_RATE && level == 1) {
			msg->body.trace.req_action_hdr.block_from = 0;
			msg->body.trace.req_action_hdr.block_until = 120;
		} else if (request == IDIP_LIMIT_RATE && level == 0) {
			msg->body.trace.req_action_hdr.block_from = 0;
			msg->body.trace.req_action_hdr.block_until = 1440;
		}
	
		return;
	}
	/*
	 * Creating a trace message is the same as a do message, just
	 * different actions. 
	 */ 
	if (type == IDIP_APP_CLASS_TRACE) {
		msg->body.trace.req_action_hdr.audit_from = 0; 
		msg->body.trace.req_action_hdr.audit_until = 10;
		return;
	}
	
	return;
	
}
/*
 * Every idip message is a idip_message_t with an embedded idip_app_msg_t.
 */
int
create_do_msg(idip_message_t *msg, char *alertmsg, char *this_host, 
    char *source, int request)
{

	uint32_t host = inet_addr(this_host);

	msg->i_hdr.next_type = IDIP_APPLICATION_DATA;
	msg->i_hdr.dest_addr = 0;
	msg->i_hdr.length = sizeof(struct idip_header);

	/*
	 * Only one option being sent, so the next type is null.
	 */
	msg->i_opt_hdr.next_type = 0;
	if (request == IDIP_LIMIT_RATE) {
		create_app_msg(&(msg->p.app_msg), IDIP_APP_CLASS_DCDO, host, 
	 	    source, alertmsg, request);
	} else if (request == IDIP_RESTORE_RATE) {
		create_app_msg(&(msg->p.app_msg), IDIP_APP_CLASS_DCUNDO, host, 
	 	    source, alertmsg, request);
	}
	return (0);
}

int
create_undo_msg(idip_message_t *msg, char *alertmsg, char *this_host, 
    char * source, int request)
{

	uint32_t host = inet_addr(this_host);

	msg->i_hdr.next_type = IDIP_APPLICATION_DATA;
	msg->i_hdr.dest_addr = 0;
	msg->i_hdr.length = sizeof(struct idip_header);
	/*
	 * Only one option being sent, so the next type is null.
	 */
	msg->i_opt_hdr.next_type = 0;

	create_app_msg(&(msg->p.app_msg), IDIP_APP_CLASS_DCUNDO, host, source,
	    alertmsg, request);
	return (0);
}

/*
 *
 * Local functions.
 */

static int
get_address(char *message, char **addr) 
{
    char *tmp = NULL;
    char *src = NULL;
    char *msg = NULL;
    char  c;

   
    if (message == NULL) {
        return (1);
    }
    msg = strdup(message);
    tmp = strtok(msg, ".");
    if(tmp == NULL) {
        return (-1);
    }

    /*
     * Now search for the first alpha character in from of this token.
     */

    for (c = *tmp; tmp != NULL; tmp++) {
	if (isdigit(c) != 0) {
	    break;
        }
    }

    /*
     * tmp should point to the first part of this address now.
     */

    fprintf(stderr, "tmp = %s now\n", tmp);
    src = malloc(strlen(tmp));
    memset(src, 0, sizeof(src));
    strncpy(src, tmp, strlen(tmp));
    fprintf(stderr, "src = %s now\n", src);

    while ((tmp = strtok(NULL, ".")) != NULL) {
	fprintf(stderr, "tmp = %s\n", tmp);
	src = realloc(src, strlen(tmp) + 1);
	src = strncat(src, ".", 1);
	fprintf(stderr, "src = %s\n",src);
	src = strncat(src, tmp, strlen(tmp));
    }

    /*
     * Now look to see if there is a trailing slash
     */

    tmp = NULL;
    if ((tmp = strchr(src, '/')) != NULL) {
	/*
	 * Take this off.
	 */
	*tmp = 0; 
    }
    *addr = strdup(src);
    return (0);
} 
