#include <string.h>

#include "idip.h"
#include "idip_message.h"
#include "nhlist.h"


/*
 * Sarah Jelinek
 * University of Colorado, Colorado Springs
 * Computer Science Masters Project.
 * 7/13/03
 * Copyright 2003
 */

static int nh_list_alloc(struct nh_list **nhl);

/*
 * Looks for the neighbor. If found, returns 0.
 * If not found returns 1. Otherwise, a non-zero value is returned in error.
 */
int 
in_nh_list(struct nh_list *nhl, struct idip_hello_entry entry)
{
	int	error = 0;
	struct nh_list *nhlp;

	if (nhl == NULL) {
		error = nh_list_alloc(&nhl);
		if (error) {
			return (error);
		}
		/* 
		 * If we had to allocate the list it isn't in there so return
		 * an error to indicate this.
		 */
		return (1);
	} 

	/*
	 * Otherwise search for it in the list
	 */
	nhlp = nhl;
	
	while (nhlp != NULL) {
		if (nhlp->nh.ip_addr == entry.id) {
			return (0);
		}
		nhlp = nhlp->next;
	}
	return (error);

}

int 
add_nh_list(struct nh_list **nhl, struct idip_hello_entry entry) 
{

	struct nh_list *tmp = *nhl;
	struct nh_list *prev;

	struct nh_data *nhp = (struct nh_data *)malloc(sizeof (struct nh_data));
	if (nhp == NULL)
		return ENOMEM;

	/* XXXX- should I do memcpy here? Probably */

	nhp->ip_addr = entry.id;
	nhp->active = entry.alive;
	nhp->type = entry.dev_type;

	prev = NULL;

	while (tmp) {
		prev = tmp;
		tmp = tmp->next;
	}

	if (prev == NULL) {
		tmp->next->nh = *nhp;
	} else {
		prev->next->nh = *nhp;
	}
	
	return 0;
}
void
set_nh_active(struct nh_data *nh) 
{
	nh->active = 1;
}
static int
nh_list_alloc(struct nh_list **nhl)
{
	*nhl = (struct nh_list *)malloc(sizeof (struct nh_list));

	if (nhl == NULL) {
		return ENOMEM;
	}

	(void) memset(*nhl, 0, sizeof (struct nh_list));
	return (0);
}
