#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>
#include <fcntl.h>

#include "idip_application.h"

/*
 * Sarah Jelinek
 * University of Colorado, Colorado Springs
 * Computer Science Masters Project.
 * 5/17/03
 * Copyright 2003
 */ 

#define IDIP_MESSAGE	  	0x0
#define	IDIP_APPLICATION_DATA	0x01
#define IDIP_HELLO_DATA		0x02
#define IDIP_NKID_DATA		0x03
#define IDIP_CRED_DATA		0x05
#define IDIP_STARTUP_DATA	0x06
#define IDIP_ESP		0x32
#define IDIP_AH			0x33

#define IDIP_DC			0x01
#define IDIP_END_SYSTEM		0x02
#define IDIP_ROUTER		0x03
#define IDIP_APP_PROXY		0x04
#define IDIP_IDS		0x08
#define IDIP_REMOTE_DC		0x32
#define IDIP_EDGE		0x64

#define IDIP_ACK		0x1
#define IDIP_ND_ACK		0x2
#define IDIP_NCI_ACK		0x4
#define IDIP_NR_ACK		0x8
#define IDIP_NN_ACK		0x10
#define IDIP_NC_ACK		0x11 

/*
 * Comm types
 */
#define IDIP_BIND		0x02
#define IDIP_RESTRICT		0x03
#define IDIP_CLOSE		0x04


struct idip_header {

	uint8_t	  version;
	uint8_t   flags;
	uint16_t  length;
	uint8_t   next_type;
	uint8_t   pad;
	uint16_t  checksum;
	uint32_t  seq_num;
	uint32_t  time_stamp;
	uint32_t  priority;
	uint32_t  dest_addr;
	uint32_t  dest_proc_id;
	uint32_t  dest_boot_time;
	uint32_t  pad_extra;
};

struct idip_option_hdr {
	uint8_t	  next_type;
	uint8_t	  length;
	uint8_t	  opt_spec_data;
};

struct idip_hello_hdr {
	uint8_t	  flags;
	uint8_t	  pad;
	uint16_t  length;
	uint16_t  crypto;
	int	  proxy;
	uint16_t  dev_type;
	uint8_t   comm_type;
	uint32_t  timestamp;
	uint32_t  id;
	uint32_t  interface1;
	uint32_t  interface2;
	uint32_t  interface3;
	uint32_t  interface4;
	uint32_t  interface5;
	uint32_t  interface6;
};

struct idip_hello_entry {
	uint32_t  id;
	uint32_t  time_delta;
	uint8_t	  is_signed;
	uint8_t	  alive;
	uint8_t	  proxy;
	uint8_t   dev_type;
	uint8_t   comm_type;
};

struct idip_mailbx_hdr {
	uint32_t  total;
	uint32_t  interface;
	uint32_t  pid;
	uint32_t  source;
};

struct idip_comm_msg {
	uint32_t  length; /* length of the payload */
	uint32_t  type;
	uint32_t  pid;
	uint32_t  boottime;
	void	 *payload; /* payload can be variable length and type */
};

typedef struct idip_hello_msg {
	struct idip_hello_hdr hello_hdr;
	struct idip_hello_entry hello_entries[16];
} idip_hello_msg_t; 

typedef struct idip_message {
	union {
		void	*generic;
		idip_app_msg_t  app_msg;
		idip_hello_msg_t hello_msg;
	} p;
	struct idip_header i_hdr;
	struct idip_option_hdr i_opt_hdr; /* list of 0 or more options */
} idip_message_t;


typedef struct idip_mbx_msg {
	struct idip_mailbx_hdr mbx_hdr;
	idip_message_t idip_msg;
} idip_mbx_msg_t;

typedef struct recv_msg {
	uint32_t    addr;
	char 	   *hostname;
	uint32_t    sid_type;
} recv_msg_t;
	
