#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <netdb.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <inttypes.h>

#include "idip_message.h"
#include "idip.h"
#include "nhlist.h"
#include "dclist.h"


static int 	init_message_protocol(struct hostent *, struct sockaddr_in *, 
			struct sockaddr_in *, struct sockaddr_in *);
static int 	to_mbxs(struct hostent *h_ent, struct sockaddr_in *ack, 
			struct sockaddr_in *rcv, struct sockaddr_in *fwd);


/*
 * Sarah Jelinek
 * University of Colorado, Colorado Springs
 * Computer Science Masters Project.
 * 5/17/03
 * Copyright 2003
 */

/*
 * The backplane listens on a socket and determines the type of request
 * being sent to it. From there it invokes the appropriate processing.
 */

void
main() {

	int	length;
	int	n;
	idip_message_t i_message;
	idip_mbx_msg_t i_mbx;
	struct	sockaddr_in ack;
	struct	sockaddr_in rcv;
	struct	sockaddr_in fwd;
	struct	sockaddr_in toaddr;
	struct	sockaddr_in thisaddr;
	struct  hostent   h_ent;
	int	error = 0;

	/*
	 * Initialize the protocol
	 */

	error = init_message_protocol(&h_ent, &ack, &rcv, &fwd);
	if (error) {
		fprintf(stderr, "Failed to initialize protocol, reason: %s \n",
			strerror(error));
		return;
	}
	if (inet_pton(AF_INET, h_ent.h_addr_list[0], 
		&thisaddr.sin_addr) <= 0) {
		perror("error translating host address\n");
		return;
	}

	/*
	 * Set up our listening socket 
	 */

	if ((gen_mbx = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
		fprintf(stderr, "Unable to set up receiver socket.\n");
		perror(strerror(errno));
		return;
	}

	(void) memset(&gen_from, 0, sizeof (gen_from));
	gen_from.sin_family = AF_INET;
	gen_from.sin_addr.s_addr = INADDR_ANY;
	gen_from.sin_port = htons(IDIP_APP_PORT);

	if (bind(gen_mbx, (struct sockaddr *) &gen_from,
		sizeof (struct sockaddr_in)) < 0) {
		fprintf(stderr, "%s", "Could not bind to port\n");
		perror(strerror(errno));
	}

	length = sizeof (gen_from);

	while (1) {
		n = recvfrom(gen_mbx,
		    &i_message,
			sizeof (struct idip_message), 0,
			    (struct sockaddr *)&toaddr, &length);
		if (n < 0) {
		    perror("receiving datagram messages");
		    continue;
		}

		/*
		 * Determine what to do next based on the following:
		 * If flag is set to ACK, then send to sender process.
		 * If flag is not set, check destination address to
		 * determine which sender process mailbox to forward
		 * this message to.
		 */

		if (i_message.i_hdr.flags != 0) {
			/* 
			 * This is an ACK message, forward it on to the
			 * ack mbx. For an ACK message, there is no payload.
			 * Send the header we received.
			 */
			memset(&i_mbx, 0, sizeof (i_mbx));
			i_mbx.mbx_hdr.total = sizeof (i_message);
			i_mbx.mbx_hdr.source = toaddr.sin_addr.s_addr;
			i_mbx.idip_msg = i_message;

			sendto(gen_mbx, (idip_mbx_msg_t *)&i_mbx,
				sizeof (idip_mbx_msg_t), 0,
				    (struct sockaddr *)&ack, 
					sizeof (ack));
		}
		/*
		 * Not an ACK, determine which mbx to send it to based on
		 * destination address.
		 */

		/*
		 * If not this host, forward it on.
		 */
		if (i_message.i_hdr.dest_addr != thisaddr.sin_addr.s_addr) {
			/*
			 * Send it on to the fwd mailbox.
			 */
			memset(&i_mbx, 0, sizeof (i_mbx));
			i_mbx.mbx_hdr.total = sizeof (i_message);
			i_mbx.mbx_hdr.source = toaddr.sin_addr.s_addr;
			i_mbx.idip_msg = i_message;

			sendto(gen_mbx, (idip_mbx_msg_t *) &i_mbx,
				sizeof (struct idip_mbx_msg), 0,
				    (struct sockaddr *)&fwd, 
					sizeof (fwd));
		} else {
			/*
			 * Send it on to the rcv mailbox.
			 */

			memset(&i_mbx, 0, sizeof (i_mbx));
			i_mbx.mbx_hdr.total = sizeof (i_message);
			i_mbx.mbx_hdr.source = toaddr.sin_addr.s_addr;
			i_mbx.idip_msg = i_message;
			sendto(gen_mbx, (struct idip_mbx_msg *) &i_mbx,
				sizeof (struct idip_mbx_msg), 0,
				    (struct sockaddr *)&rcv, 
					sizeof (rcv));
		}
	}

	(void) close(gen_mbx);
}
/*
 * Set up mbx's that are used to communiate with the sender process.
 */
static int
to_mbxs(struct hostent *h_ent, struct sockaddr_in *ack, 
	struct sockaddr_in *rcv, struct sockaddr_in *fwd)
{

	(void) memset(ack, 0, sizeof (*ack));
	ack->sin_family = AF_INET;
	inet_pton(AF_INET, h_ent->h_addr_list[0],
	    &ack->sin_addr);
	ack->sin_port = htons(2004);

	(void) memset(rcv, 0, sizeof (*rcv));
	rcv->sin_family = AF_INET;
	inet_pton(AF_INET, h_ent->h_addr_list[0],
	    &rcv->sin_addr);
	rcv->sin_port = htons(2005);

	(void) memset(fwd, 0, sizeof (*fwd));
	fwd->sin_family = AF_INET;
	inet_pton(AF_INET, h_ent->h_addr_list[0],
	    &fwd->sin_addr);
	fwd->sin_port = htons(2006);

	return (0);

}

static int
init_message_protocol(struct hostent *h_ent, struct sockaddr_in *ack, 
	struct sockaddr_in *rcv, struct sockaddr_in *fwd)
{
	int	error = 0;
	char	*argv[1];

	argv[0] = "idip_msg_sender";

	/*
	 * Start the sender process
	 */

/*	error = execl("./idip_msg_sender", (const char *) argv[0]);
	if (error == -1) {
		return (errno);
	}*/ 
	/*
	 * Setup the socket address that are used to communicate with the
	 * sender process.
	 * Get this host entry to use to send these messages.
	 */
	h_ent = gethostent();
	to_mbxs(h_ent, ack, rcv, fwd);
	(void) endhostent();

	return (0);
}
