#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <string.h>
#include <strings.h>
#include <errno.h>
#include <unistd.h>
#include <netdb.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <inttypes.h>
#include <stropts.h>
#include <sys/select.h>

#include "idip_message.h"
#include "idip.h"
#include "nhlist.h"
#include "dclist.h"
#include "acklist.h"

/*
 * Sarah Jelinek
 * University of Colorado, Colorado Springs
 * Computer Science Masters Project.
 * 5/17/03
 * Copyright 2003
 */

static int init_sender_process(struct hostent *h_ent);

static int send_multicast_msg(struct sockaddr_in from, idip_message_t msg); 

static int process_send_msg(struct hostent *h_ent, idip_message_t msg, 
		struct sockaddr_in who, struct sockaddr_in fwd_addr);

static int process_ack_msg(int ack, idip_message_t msg, 
	struct sockaddr_in who);

static int process_rcv_msg(idip_message_t msg, 
	struct sockaddr_in who);

static int process_fwd_msg(int fwd, idip_message_t msg, 
	struct sockaddr_in who);

static int process_hello_msg(idip_message_t *);

static int process_dc_data(struct idip_hello_entry entry);

static int process_nl_data(struct idip_hello_entry entry);

static int notify_neighbors(struct idip_hello_entry entry);

/*
 * global mailboxes(sockets) which to listen on
 */
 
static int	fwd;
static int	rcv;
static int	snd;
static int	ack;

static struct sockaddr_in fwd_addr;
static struct sockaddr_in rcv_addr;
static struct sockaddr_in snd_addr;
static struct sockaddr_in ack_addr;

static struct hostent	  *h_ent;
/*
 * Host list for registration of messages.
 */
recv_msg_t rmsg_undo[100];
recv_msg_t rmsg_trace[100];


void
main()
{

	int	error;
	int	n;
	int	num = 0;
	int	length = 0;
	struct	sockaddr_in	who;
	idip_message_t		send_msg;
	idip_message_t		ack_msg;
	idip_message_t		fwd_msg;
	idip_message_t		rcv_msg;

	fd_set			rset;
	fd_set			wset;
	 

	/*
	 * Initialize the mailboxes for the receiver to communication with
	 * us
	 */

	sethostent(0);
	h_ent = gethostent();
	endhostent();

	error = init_sender_process(h_ent);
	if (error) {
		fprintf(stderr, "Failed to initialize sender process, \
			reason: %s\n", strerror(error));
			return;
	}

	FD_ZERO(&rset);
	FD_ZERO(&wset);

	FD_SET(snd, &rset);
	FD_SET(rcv, &rset);
	FD_SET(fwd, &rset);
	FD_SET(ack, &rset);

	FD_SET(snd, &wset);
	FD_SET(rcv, &wset);
	FD_SET(fwd, &wset);
	FD_SET(ack, &wset);

	while (1) {
		num = select(4, &rset, &wset, NULL, 0);
		if (num > 0) {
			if (FD_ISSET(snd, &rset)) {
				/*
				 * Process send message.
				 */

				n = recvfrom(snd, &send_msg, 
					sizeof (idip_message_t), 0,
						(struct sockaddr *)&who, 
						&length);

				process_send_msg(h_ent, send_msg, who,
					fwd_addr);
			}
			if (FD_ISSET(rcv, &rset)) {
				n = recvfrom(rcv, &rcv_msg, 
					sizeof (idip_message_t), 0, 
						(struct sockaddr *)&who, 
						&length);

				process_rcv_msg(rcv_msg, who);
			}

			if (FD_ISSET(ack, &rset)) {
				n = recvfrom(ack, &ack_msg, 0, 
					sizeof (idip_message_t), 
						(struct sockaddr *)&who, 
						&length);

				process_ack_msg(ack, ack_msg, who);
			}
			if (FD_ISSET(fwd, &rset)) {
				n = recvfrom(fwd, &fwd_msg, 0, 
					sizeof (idip_message_t), 
						(struct sockaddr *)&who, 
						&length);

				process_fwd_msg(fwd, fwd_msg, who);
			}
		}
	}
			
		
}

/*
 * Initialize all of the mailboxes that the sender should be
 * listening to
 *	send mbx
 *	rcv mbx
 *	ack mbx
 *	fwd mbx
 *	gen mbx	
 */
static int
init_sender_process(struct hostent *h_ent)  
{



	(void) memset(&snd_addr, 0, sizeof (snd_addr));
        snd_addr.sin_family = AF_INET;
	inet_pton(AF_INET, h_ent->h_addr_list[0], &snd_addr.sin_addr);
        snd_addr.sin_port = htons(2000);

	(void) memset(&ack_addr, 0, sizeof (ack_addr));
        ack_addr.sin_family = AF_INET;
	inet_pton(AF_INET, h_ent->h_addr_list[0], &ack_addr.sin_addr);
        ack_addr.sin_port = htons(2001);

        (void) memset(&rcv_addr, 0, sizeof (rcv_addr));
        rcv_addr.sin_family = AF_INET;
	inet_pton(AF_INET, h_ent->h_addr_list[0], &rcv_addr.sin_addr);
        rcv_addr.sin_port = htons(2002);

        (void) memset(&fwd_addr, 0, sizeof (fwd_addr));
        fwd_addr.sin_family = AF_INET;
	inet_pton(AF_INET, h_ent->h_addr_list[0], &fwd_addr.sin_addr);
        fwd_addr.sin_port = htons(2003);

	if ((snd = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
		return (errno);
	}

	if (bind(snd, (struct sockaddr *)&snd_addr, 
		sizeof (struct sockaddr_in)) < 0) {
		return (errno);
	}

	if ((rcv = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
		return (errno);
	}

	if (bind(rcv, (struct sockaddr *)&rcv_addr, 
		sizeof (struct sockaddr_in)) < 0) {
		return (errno);
	}
	if ((ack = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
		return (errno);
	}

	if (bind(ack, (struct sockaddr *)&ack_addr, 
		sizeof (struct sockaddr_in)) < 0) {
		return (errno);
	}
	if ((fwd = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
		return (errno);
	}

	if (bind(fwd, (struct sockaddr *)&fwd_addr, 
		sizeof (struct sockaddr_in)) < 0) {
		return (errno);
	}
}
/*
 * Send messages based on the following:
 *	-If the destination is the localhost, then this sender
 *	process delivers the message to all of those applications
 *	that have registered for delivery of this type of message
 *
 *	-If the destination is not the localhost, then this message
 *	gets forwarded as is on to the destination
 */


static int
process_send_msg(struct hostent *h_ent, idip_message_t msg, 
	struct sockaddr_in who, struct sockaddr_in fwd_addr)
{

	int	error = 0;
	/*
	 * Check to see if the destination is this host. If so, deliver
	 * this message to all applications that have registered for delivery
	 * of this kind of message. If not, forward the whole thing on to
	 * the destination.
	 */

	if (msg.i_hdr.dest_addr == inet_addr(h_ent->h_addr_list[0])) {
		/*
		 * Send this message to those on the notification list
		 * for this type of message.
		 */

		error = send_multicast_msg(who, msg);
		if (error) {
			fprintf(stderr, 
			    "Unable to send multicast message: %s\n", 
				strerror(errno));
				return (error);
		}
			
	} else {
		/*
		 * Forward this to the destination. Use the fwd mailbox
		 * to handle this.
		 */
		if (sendto(fwd, (idip_message_t *)&msg, sizeof(idip_message_t),
			0, (struct sockaddr *)&fwd_addr,  
			    sizeof(fwd_addr)) < 0) {
				perror("Could not fwd message");
				return (-1);
		}
	}
	
	return (0);
}

static int
process_ack_msg(int ack, idip_message_t msg, struct sockaddr_in who)
{
	int	error = 0;
	/*
	 * An ACK message is marked in the pending ACK list and the
	 * entry is cleared.
	 */
	
	error = remove_ack_list(&ackl, who.sin_addr.s_addr);
	return (error);
}

/*
 * Processes all messages in rcv mailbox as follows:
 * 	Looks at type of message request. Processes based on
 * 	type, either do rate limiting, or trace packets.
 */ 
static int
process_rcv_msg(idip_message_t msg, struct sockaddr_in who)
{

	struct sockaddr_in	trace_sock;
	if (msg.p.app_msg.app_hdr.class_id == IDIP_APP_CLASS_TRACE) {
		/*
		 * Send this request to the tracelistener on this host.
		 */
		if (sendto(gen_mbx, (idip_message_t *)&msg, sizeof (msg), 0,
		    (struct sockaddr *)&trace_sock, sizeof (trace_sock)) < 0) {
			perror("trace mbx");
			return (-1);
		}
	
	}	
	return (0);
}
/*
 * The sender process forwards this message on to the destination, or to
 * the next hop to the destination
 */
static int
process_fwd_msg(int fwd, idip_message_t msg, struct sockaddr_in who)
{

	/*
	 * Do no processing of this message. Just forward it on to
	 * the destination specified.
	 */
	
	bcopy(&msg.i_hdr.dest_addr, &gen_from.sin_addr, 
	    sizeof (gen_from.sin_addr)); 

	if (sendto(gen_mbx, (idip_message_t *) &msg, sizeof (idip_message_t),
		0, (struct sockaddr *) &gen_from, sizeof (gen_from)) < 0) {
		    perror("processing fwd message");
		    return (-1); 
	}

	return (0);
}

static int
send_multicast_msg(struct sockaddr_in from, idip_message_t msg)
{

	int has_multi = 0;
	int i;
	int type;

	/*
	 *  At this time we only send out application data requests
	 *  for Do/Undo and trace messages.
	 */
	if (msg.i_hdr.next_type == IDIP_APPLICATION_DATA) {
		type = (int) msg.p.app_msg.app_hdr.class_id; 
	} else {
	    return (EINVAL);
	}

	/*
	 * If multicast available, send this way. If not, send
	 * unicast, one by one.
	 */
	if (has_multi) {
	} else {
	    switch (type) {
		case IDIP_APP_CLASS_TRACE :
		case IDIP_APP_CLASS_DCDO :
		    for (i = 0; i < sizeof(rmsg_trace); i ++) {
			bcopy(&rmsg_trace[i].addr, 
				&gen_from.sin_addr, sizeof(from.sin_addr));
			if (sendto(gen_mbx, 
				(struct idip_app_trace_msg *)&msg.p.app_msg.body.trace,
				    sizeof(struct idip_app_trace_msg), 0,
				        (struct sockaddr *)&gen_from, 
					    sizeof(gen_from)) < 0) {
					perror("sending multicast message");
					return (-1);
			}
		    }
		    break;

		/*
	 	 * Registration for a Do assumes a registration for an
		 * Undo.
		 */
		case IDIP_APP_CLASS_DCUNDO :
		    for (i = 0; i < sizeof(rmsg_undo); i ++) {
			bcopy(&rmsg_undo[i].addr, 
				&gen_from.sin_addr, sizeof(from.sin_addr));
			if (sendto(gen_mbx, 
			    (struct idip_app_dc_undo_msg *)&msg.p.app_msg.body.undo, 
				sizeof(struct idip_app_dc_undo_msg), 0,
				(struct sockaddr *)&gen_from, 
					sizeof(gen_from)) < 0) {
					perror("sending multicast message");
					return (-1);
			}
		    }
		    break;
	    }
		
	}
	return (0);
}

/*
 * Handles all hello messages as follows:
 *	-If this is a hello message that indicates the source is a DC node,
 *	this node is added to DC list, if not already included
 * 	-If DC hello message, the current node sets itself as a proxy to
 *	that DC and notifies all of the neighbors of this status
 *	-If not DC, then this node checks to see if source node is
 *	a neighbor. If it is, it does two things:
 *		-Modifes the record to show that the source node is an
 *		active neighbor
 *		-Refreshes the timing values for the node
 *		-Updates the source nodes neighborhood list
 *	-If source node not in neighbor list, it is added and the neighbors list
 *		-Then sends a message to current node neighborhood to
 *		update their neighbor lists
 */ 
  
static int
process_hello_msg(idip_message_t *msg) {

	idip_hello_msg_t	hmsg;
	int			error = 0;
	int			i = 0;

	hmsg = msg->p.hello_msg;

	/*
	 * Process all of the hello entries.
	 */

	for (i = 0; i < sizeof (hmsg.hello_entries); i ++) {

	    /* 
	     * If the message is not the right size, or has no payload
	     * return an error.
	     */
	    if ((sizeof (hmsg) != sizeof (idip_hello_msg_t))
		|| (hmsg.hello_hdr.length == 0)) {
		return (EINVAL);
	    }
	    if (hmsg.hello_hdr.dev_type & IDIP_DC) {
		if ((error = process_dc_data(hmsg.hello_entries[i])) 
			!= 0)
		    return (error);
	    } else {
		if ((error = process_nl_data(hmsg.hello_entries[i])) != 0)
		    return (error);
	    	
	    }
	}
	return (error);
	
}

/*
 * Adds data to DC list. Notifies neighbors of new proxy status
 */
static int
process_dc_data(struct idip_hello_entry entry) {

	int			error = 0;

	if (!in_dc_list(dcl, entry)) {
	    error = add_dc_list(&dcl, entry);
	}
	/*
	 * If successfully added to dc list, notify neighbors. Otherwise
  	 * do not.
	 */
	if (!error) {
		return (error = notify_neighbors(entry));
	}

	return (error);
}

/*
 * Adds node to neighborhood list if not already an entry.
 * Updates the active status of the neighbor
 */
static int
process_nl_data(struct idip_hello_entry entry) {

	int	error = 0;

	if (!in_nh_list(nhl, entry)) {
		return (error = add_nh_list(&nhl, entry));
	}
	return (error);
}
/*
 * Notifies this nodes neighbors of the udpated proxy status for the 
 * the specified DC node.
 */
static int
notify_neighbors(struct idip_hello_entry entry)
{
	int	error = 0;
	int	i = 0;
	struct idip_hello_entry proxy_entry;

	uint32_t haddr = inet_addr(h_ent->h_addr_list[0]);

	proxy_entry.id = haddr;
	proxy_entry.alive = 1;
	proxy_entry.proxy = 1;


	for (i = 0; nhl->next != NULL; i ++) {
	    /*
	     * Send hello messages to all the neighbors in the list,
	     * with the indication that this node is a proxy to that DC.
	     */

	     bcopy(&nhl[i].nh.ip_addr, &gen_from.sin_addr, 
		sizeof (gen_from.sin_addr));
	     if (sendto(gen_mbx, (struct idip_hello_entry *) &proxy_entry, 
		sizeof (struct idip_hello_entry), 0, (struct sockaddr *)
			&gen_from, sizeof (gen_from)) < 0) {
			perror("processing notify neighbors");
			return (-1);
	    }
	}
		
	return (error);
}

