/*
 * Sarah Jelinek
 * University of Colorado, Colorado Springs
 * Masters of Science, Computer Science project
 * February 14, 2004
 * 
 * Deamon program that automatically send the appropriate IDIP Message to the 
 * IDIP Message Layer application listening on port 0xc1df. 
 *                                     -------------------------
 */

/*
 * The spo_alert_unixsock.c file the current version of snort doesn't
 * attach information into alert. This code will create the appropriate IDIP
 * message to send to the IDIP Message Layer application.
 * 
 *  //  we don't log any headers besides eth yet 
 *     if (p->iph) alertpkt.nethdr=(char *)p->iph-(char *)p->pkt;
 *     
 *       switch(p->iph->ip_proto)
 *       {
 *           case IPPROTO_TCP:
 *               if (p->tcph) alertpkt.transhdr=(char *)p->tcph-(char *)p->pkt;
 *               break;
 *
 *           case IPPROTO_UDP:
 *               if (p->udph) alertpkt.transhdr=(char *)p->udph-(char *)p->pkt;
 *               break;
 *
 *           case IPPROTO_ICMP:
 *               if (p->icmph) alertpkt.transhdr=(char *)p->icmph-(char *)p->pkt;
 *               break;
 *
 *           default:
 *               // alertpkt.transhdr is null due to initial bzero 
 *               alertpkt.val|=NO_TRANSHDR;
 *       }
 *       if (p->data) alertpkt.data=p->data - p->pkt;
 *   }
 * 
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <signal.h>
#include "snort.h"
#include "msock.h"
#include "idip.h"
#include "idip_message.h"
#include "idip_application.h"
#include "message.h" 

int debug = 0;  /* debug flag */

FILE *logfp = (FILE *)NULL;

int 
analyzePacket(Alertpkt *alert, FILE *log, char *hostname, char *messagehost,
	int portno)
{
    Packet 	   *p;
    char 	   ipaddr[50];
    int 	   clientfd;
    int		   messagefd;
    time_t 	   now;
    char 	   *time_string;

    struct idip_app_trace_msg idip_trace_msg; /* request for tracing a source */
    struct idip_app_trace_msg idip_do_msg; /* request for rate limiter */  


    memset (&idip_trace_msg, 0, sizeof (struct idip_app_trace_msg)); 
    memset (&idip_do_msg, 0, sizeof (struct idip_app_trace_msg)); 

    if (!(alert->val & NOPACKET_STRUCT)) {
	if ((p = calloc (1, sizeof (Packet))) == NULL) {
	    fprintf(stderr, 
		"unable to allocate memory for the packet structure");
	    return;
	}
      
	p->pkt = alert->pkt;
	p->pkth = &alert->pkth;
	if (alert->dlthdr)
		p->eh = (EtherHdr *) (alert->pkt + alert->dlthdr);
	if (alert->nethdr) {
		p->iph = (IPHdr *) (alert->pkt + alert->nethdr);
		if (alert->transhdr) {
		    switch (p->iph->ip_proto) {
			case IPPROTO_TCP:
			    p->tcph = (TCPHdr *) (alert->pkt + alert->transhdr);
			    break;
			case IPPROTO_UDP:
			    p->udph = (UDPHdr *) (alert->pkt + alert->transhdr);
			    break;
			case IPPROTO_ICMP:
			    p->icmph = (ICMPHdr *) (alert->pkt + 
				alert->transhdr);
			    break;
			default:
			    fprintf (stdout, "need to analyse the protcol 
				%d !\n",p->iph->ip_proto);
                    }
	    }
	}
      
	if (alert->data)
	    p->data = alert->pkt + alert->data;
    } 

    printf("alert->val:%d NO_PACKET_STRUCT:%d 
	(alert->val & NOPACKET_STRUCT):%d\n", alert->val, 
	    NOPACKET_STRUCT, (alert->val & NOPACKET_STRUCT));
 
    now = time(NULL);
    time_string = ctime(&now);
    time_string[strlen(time_string) - 1] = '\0';
   
    fprintf (log,"[%s] %s [%d]\n", time_string, alert->alertmsg, 
	alert->event.event_id);

    if (alert->val) {
	if ((messagefd = connectTo(messagehost, portno)) == -1) {
	    (void) fprintf(stderr,"Failed to connect to message layer \n");
	    return -1;
	}
	/*
	* Generate the appropriate IDIP Trace and forward request to send to 
	* message layer node. This IDIP message layer will send a trace request
	* all other idip nodes in the neighborhood. The Message Layer will also
	* forward the forward request to the appropriate IDIP node for 
	* processing.
	*/
	if (create_trace_msg(&idip_trace_msg, alert) != 0) {
		fprintf(stderr, "Failed to create trace request \n");
		return (-1);
	}

	/* send the request to the IDIP Message Layer  */
	sockWrite(messagefd, idip_trace_msg, sizeof(idip_app_trace_msg));

	if (create_do_msg(&idip_do_msg, alert, hostname) != 0) {
	    fprintf(stderr, "Failed to create do request \n");
	    return (-1);
	}
	sockWrite(messagefd, idip_do_msg, sizeof(idip_app_do_msg));
    }
  
    if (!(alert->val & NOPACKET_STRUCT)) {
	  
	if (p->iph && (p->tcph || p->udph || p->icmph)) {
	    switch (p->iph->ip_proto) {
		case IPPROTO_TCP:
		    fprintf(log, "TCP from: %s:%d ",inet_ntoa(p->iph->ip_src), 
			ntohs(p->tcph->th_sport));
		    fprintf(log,"to: %s:%d\n", inet_ntoa(p->iph->ip_dst), 
			ntohs(p->tcph->th_dport));
		    break;
		case IPPROTO_UDP:
		    fprintf(log,"UDP from: %s:%d ", inet_ntoa(p->iph->ip_src), 
			ntohs (p->udph->uh_sport));
		    fprintf(log,"to: %s:%d\n", inet_ntoa (p->iph->ip_dst), 
			ntohs (p->udph->uh_dport));
		    break;
		case IPPROTO_ICMP:
		    fprintf (log,"ICMP type: %d code: %d from: %s ", 
			p->icmph->type, p->icmph->code, 
			    inet_ntoa (p->iph->ip_src));
			fprintf (log,"to: %s\n", inet_ntoa (p->iph->ip_dst));
			break;
	    }
	}
    }
  
    close(clientfd);
    close(messagefd);
    fflush(log);
    return (1);
}

void help_msg(char *prog) 
{

    fprintf(stdout,"Usage : %s [-d ] -h <hostname> -p <port no> -l  \
	<log file name> \n", prog);
    fprintf(stdout, " option d	turn on the debugging mode \n");
    fprintf(stdout, " deafult port no is %d \n",PORTNO);
    return;
}

void 
main (int argc, char *argv[])
{
    struct sockaddr_un  bogus;
    Alertpkt 		alert;
    char 		*host;
    char 		*logfile;
    int  		portno = 0;

    socklen_t 		len = sizeof (struct sockaddr_un);
    int 		sockfd;
    int 		c;
    struct 		sockaddr_un snortaddr;
 
    while ((c = getopt(argc, argv, "dhpl")) != EOF) {
	switch(c) {
	    case 'd':
		debug = 1;
		fprintf(stdout, "debugging mode turned on \n");
		break;
            case 'h':
		host = argv[optind];
		if (debug)
		    fprintf(stdout, "hostname : %s \n",host);
		break;
	    case 'p':
		portno = atoi(argv[optind]);
		if (debug)
		    fprintf(stdout, "port # %d \n", portno);
		break;
	    case 'l':
		logfile = argv[optind];
	        if (debug) 
		    fprintf(stdout," log file is %s \n",logfile);
		break;
	    default:
		fprintf(stdout, "unrecognized arg >%s< %d \n", optarg, optind);
		help_msg(argv[0]);
		return (-1);
	}
    }
    /* open the log file */
    if ((logfp = fopen("tmp.log","a+")) == NULL) {
	(void) fprintf(stderr, "Unable to open the log file for logging");
	return -1;
    }

    /* open the socket for checking the connections */

    unlink(UNSOCK_FILE);
  
    if ((sockfd = socket(AF_UNIX, SOCK_DGRAM, 0)) < 0) {
	perror("socket");
	exit (1);
    }

    bzero (&snortaddr, sizeof (snortaddr));
    snortaddr.sun_family = AF_UNIX;
    strcpy (snortaddr.sun_path, UNSOCK_FILE);

    if (bind (sockfd, (struct sockaddr *) &snortaddr, 
	sizeof (snortaddr)) < 0) {
	perror ("bind");
	exit (1);
    }
	    

    /* check if any alert have been received */
    while (recvfrom(sockfd, (void *)&alert, sizeof (alert), 0, 
	(struct sockaddr *) &bogus, &len) > 0) {
	if (analyzePacket(&alert, logfp, host, IDIP_APP_PORT) < 0) {
	    fprintf(stderr, "Unable to respond to alert \n");
	    perror ("recvfrom");
	    break; 
	}
    }
 
    fclose(logfp);
    close(sockfd);
    unlink (UNSOCK_FILE);
    return (1);
}
