/*
 * copyrights of this program is held by snort. 
 * 
 * Deamon program that automatically send the information to the firewall deamon 
 * so that the firewall scripts can be autonomously.
 *                                     -------------------------
 *                                     written by Angela Cearns
 *                                        acearns[at]yahoo.com
 *  
 *  Inputs & support provided by gkgodava[at]uccs.edu
 */

/*
 * I looked into the spo_alert_unixsock.c file the current version of snort doesn't
 * attach packet information into alert. i am using the code for future use -- i guress
 * 
 *  //  we don't log any headers besides eth yet 
 *     if (p->iph) alertpkt.nethdr=(char *)p->iph-(char *)p->pkt;
 *     
 *       switch(p->iph->ip_proto)
 *       {
 *           case IPPROTO_TCP:
 *               if (p->tcph) alertpkt.transhdr=(char *)p->tcph-(char *)p->pkt;
 *               break;
 *
 *           case IPPROTO_UDP:
 *               if (p->udph) alertpkt.transhdr=(char *)p->udph-(char *)p->pkt;
 *               break;
 *
 *           case IPPROTO_ICMP:
 *               if (p->icmph) alertpkt.transhdr=(char *)p->icmph-(char *)p->pkt;
 *               break;
 *
 *           default:
 *               // alertpkt.transhdr is null due to initial bzero 
 *               alertpkt.val|=NO_TRANSHDR;
 *       }
 *       if (p->data) alertpkt.data=p->data - p->pkt;
 *   }
 * 
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <signal.h>
#include "snort.h"
#include "msock.h"

#define PORTNO 1234

int debug = 0;  /* debug flag */

FILE *logfp = (FILE *) NULL;

int analyzePacket ( Alertpkt *alert, FILE *log, char *hostname, int portno)
{
  Packet *p;
  char ipaddr[50];
  int clientfd;
  time_t now;
  char *time_string;

  if (!(alert->val & NOPACKET_STRUCT))
    {
      if ((p = calloc (1, sizeof (Packet))) == NULL)
	{
	  fprintf(stderr," unable to allocate memory for the packet structure");
         return;
	}
      
      p->pkt = alert->pkt;
      p->pkth = &alert->pkth;
      if (alert->dlthdr)
	p->eh = (EtherHdr *) (alert->pkt + alert->dlthdr);
      if (alert->nethdr)
        {
          p->iph = (IPHdr *) (alert->pkt + alert->nethdr);
          if (alert->transhdr)
	    {
             switch (p->iph->ip_proto)
	       {
	       case IPPROTO_TCP:
		 p->tcph = (TCPHdr *) (alert->pkt + alert->transhdr);
		 break;
	       case IPPROTO_UDP:
		 p->udph = (UDPHdr *) (alert->pkt + alert->transhdr);
		 break;
	       case IPPROTO_ICMP:
		 p->icmph = (ICMPHdr *) (alert->pkt + alert->transhdr);
		 break;
	       default:
		 fprintf (stdout, "need to analyse the protcol %d !\n",p->iph->ip_proto);
                    }
	    }
	}
      
      if (alert->data)
	p->data = alert->pkt + alert->data;
    } 

   printf("alert->val:%d NO_PACKET_STRUCT:%d (alert->val & NOPACKET_STRUCT):%d\n", alert->val,NOPACKET_STRUCT,(alert->val & NOPACKET_STRUCT));
 
   now = time(NULL);
   time_string = ctime(&now);
   time_string[strlen(time_string) - 1] = '\0';
   
   
   
  fprintf (log,"[%s] %s [%d]\n", time_string, alert->alertmsg, alert->event.event_id);
  

  if (alert->val)
  {
	if((clientfd=connectTo(hostname, portno)) == -1) {
	  (void) fprintf(stderr,"Failed to connect to rate limiter \n");
     return -1;
	}
	
	/* send the request to the Rate Limiter */
	sockWrite(clientfd,alert->alertmsg,strlen(alert->alertmsg));
  }

  
  if (!(alert->val & NOPACKET_STRUCT))
  {
	  
    if (p->iph && (p->tcph || p->udph || p->icmph))
      {
	switch (p->iph->ip_proto)
	  {
	  case IPPROTO_TCP:
	    fprintf(log, "TCP from: %s:%d ",inet_ntoa(p->iph->ip_src), ntohs(p->tcph->th_sport));
	    fprintf(log,"to: %s:%d\n", inet_ntoa(p->iph->ip_dst), ntohs(p->tcph->th_dport));
	    break;
	  case IPPROTO_UDP:
                fprintf(log,"UDP from: %s:%d ", inet_ntoa(p->iph->ip_src), ntohs (p->udph->uh_sport));
                fprintf(log,"to: %s:%d\n", inet_ntoa (p->iph->ip_dst), ntohs (p->udph->uh_dport));
                break;
	  case IPPROTO_ICMP:
	    fprintf (log,"ICMP type: %d code: %d from: %s ", p->icmph->type, p->icmph->code, inet_ntoa (p->iph->ip_src));
	    fprintf (log,"to: %s\n", inet_ntoa (p->iph->ip_dst));
	    break;
	  }
      }
    }
  
  close(clientfd);
  fflush(log);
  return 1;
}

void help_msg(char *prog)
{
  fprintf(stdout,"Usage : %s [-d ] -h <hostname> -p <port no> -l <log file name> \n", prog);
  fprintf(stdout, " option d	turn on the debugging mode \n");
  fprintf(stdout, " deafult port no is %d \n",PORTNO);
  return;
}

int main (int argc, char *argv[])
{
 struct sockaddr_un bogus;
 Alertpkt alert;
 char *host;
 char *logfile;
 int  portno=0;
 socklen_t len = sizeof (struct sockaddr_un);
 int sockfd;
 int c;
 struct sockaddr_un snortaddr;
 
 while ((c=getopt(argc,argv,"dhpl")) != EOF)
   {
     switch(c)
       {
       case 'd':
	 debug = 1;
	 if (debug) 
	   fprintf(stdout,"debugging mode turned on \n");
	 break;
           case 'h':
	     host = argv[optind];
	     if (debug)
	       fprintf(stdout,"hostname : %s \n",host);
	     break;
       case 'p':
	 portno = atoi(argv[optind]);
	 if (debug)
			     fprintf(stdout,"port # %d \n",portno);
	 break;
       case 'l':
	 logfile = argv[optind];
	 if (debug) 
			     fprintf(stdout," log file is %s \n",logfile);
	 break;
       default:
	 fprintf(stdout,"unrecognized arg >%s< %d \n", optarg, optind);
	 help_msg(argv[0]);
	 return (-1);
       }
  }
 /* open the log file */
 if ( (logfp = fopen("tmp.log","a+")) == NULL)
   {
	(void) fprintf(stderr, "Unable to open the log file for logging");
	return -1;
   }


 /* open the socket for checking the connections */

  unlink (UNSOCK_FILE);
  
  if ((sockfd = socket (AF_UNIX, SOCK_DGRAM, 0)) < 0)
   {
     perror ("socket");
     exit (1);
   }

   bzero (&snortaddr, sizeof (snortaddr));
   snortaddr.sun_family = AF_UNIX;
   strcpy (snortaddr.sun_path, UNSOCK_FILE);

  if (bind (sockfd, (struct sockaddr *) &snortaddr, sizeof (snortaddr)) < 0)
   {
     perror ("bind");
     exit (1);
   }
	    

 /* check if any alert have been received */
 while (recvfrom (sockfd, (void *) &alert, sizeof (alert), 0, (struct sockaddr *) &bogus, &len) > 0)
    { 

	    printf("got a request yahoooo.........!!!\n");
      if (analyzePacket ( &alert, logfp, host, portno) < 0)
	break; 
    }
 
 perror ("recvfrom");
 fclose(logfp);
  close(sockfd);
  unlink (UNSOCK_FILE);
  
  return 1;
}


