/*
 * Sarah Jelinek
 * University of Colorado, Colorado Springs
 * Masters of Science, Computer Science project
 * February 14, 2004
 * 
 * Deamon program that automatically send the appropriate IDIP Message to the 
 * IDIP Message Layer application listening on port 0xc1df. 
 *                                     -------------------------
 */

/*
 * The spo_alert_unixsock.c file the current version of snort doesn't
 * attach information into alert. This code will create the appropriate IDIP
 * message to send to the IDIP Message Layer application.
 * 
 *  //  we don't log any headers besides eth yet 
 *     if (p->iph) alertpkt.nethdr=(char *)p->iph-(char *)p->pkt;
 *     
 *       switch(p->iph->ip_proto)
 *       {
 *           case IPPROTO_TCP:
 *               if (p->tcph) alertpkt.transhdr=(char *)p->tcph-(char *)p->pkt;
 *               break;
 *
 *           case IPPROTO_UDP:
 *               if (p->udph) alertpkt.transhdr=(char *)p->udph-(char *)p->pkt;
 *               break;
 *
 *           case IPPROTO_ICMP:
 *               if (p->icmph) alertpkt.transhdr=(char *)p->icmph-(char *)p->pkt;
 *               break;
 *
 *           default:
 *               // alertpkt.transhdr is null due to initial bzero 
 *               alertpkt.val|=NO_TRANSHDR;
 *       }
 *       if (p->data) alertpkt.data=p->data - p->pkt;
 *   }
 * 
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/un.h>
#include "snort.h"
#include "../../idip/idip_include/message.h" 
#include "../../idip/idip_include/idip.h"


int debug = 0;  /* debug flag */

FILE *logfp = (FILE *)NULL;

int 
analyzePacket(Alertpkt *alert, FILE *log, char *hostname, int portno)
{
	Packet	*p = NULL; /* This is used for logging data only */
	time_t	now;
	int	snd;
	int	created_trace = 0;
	char	*time_string = NULL;
	struct	sockaddr_in	snd_addr;
	idip_message_t idip_trace_msg; /* request for tracing a source */
	idip_message_t idip_do_msg; /* request for rate limiter */
	in_addr_t	haddr;

	memset (&idip_trace_msg, 0, sizeof (idip_message_t)); 
	memset (&idip_do_msg, 0, sizeof (idip_app_msg_t)); 

	if ((snd = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
		fprintf(stderr, "Unable to setup sender socket \n");
		perror(strerror(errno));
		return (-1);
	}
if (debug)
fprintf(stderr, "socket opened\n");

	haddr = inet_addr("192.168.13.99");
	memset(&snd_addr, 0, sizeof (snd_addr));
	snd_addr.sin_family = AF_INET;
	bcopy(&haddr, &snd_addr.sin_addr.s_addr, sizeof (haddr));
	snd_addr.sin_port = htons(IDIP_APP_PORT);

	if (!(alert->val & NOPACKET_STRUCT)) {
		if ((p = calloc (1, sizeof (Packet))) == NULL) {
			fprintf(stderr, "unable to allocate memory "
			    " for the packet structure");
			return (-1);
		}
      
		p->pkt = alert->pkt;
		p->pkth = &alert->pkth;
		if (alert->dlthdr)
			p->eh = (EtherHdr *) (alert->pkt + alert->dlthdr);
		if (alert->nethdr) {
			p->iph = (IPHdr *) (alert->pkt + alert->nethdr);
			if (alert->transhdr) {
				switch (p->iph->ip_proto) {
				case IPPROTO_TCP:
					p->tcph = (TCPHdr *) (alert->pkt + 
					    alert->transhdr);
					break;
				case IPPROTO_UDP:
					p->udph = (UDPHdr *) (alert->pkt + 
					    alert->transhdr);
					break;
				case IPPROTO_ICMP:
if (debug)
fprintf(stderr, "icmp packet found \n");
					p->icmph = (ICMPHdr *) (alert->pkt + 
					    alert->transhdr);
					break;
				default:
					fprintf (stdout, "need to analyze "
					    "the protcol %d !\n", 
					    p->iph->ip_proto);
				}
			}
		}
      
		if (alert->data)
			p->data = alert->pkt + alert->data;
	} 

if (debug)
printf("alert->val:%d NO_PACKET_STRUCT:%d (alert->val & NOPACKET_STRUCT):%d\n",
 alert->val,NOPACKET_STRUCT,(alert->val & NOPACKET_STRUCT));

	now = time(NULL);
	time_string = ctime(&now);

if (debug) 
fprintf(stderr, "time string = %s\n", time_string);

	fprintf (log,"[%s] %s [%d]\n", time_string, alert->alertmsg, 
	   alert->event.event_id);

	/*
	 * Generate the appropriate IDIP Trace and forward request 
	 * to the identified source firewall address. The ip_src
	 * address is the address of the host from which the attack
	 * has been detected.
	 */
	fflush(stderr);

	if ((p != NULL) && (strstr((char *)alert->alertmsg, "End") == NULL)) {
		if (create_trace_msg(&idip_trace_msg, (char *)alert->alertmsg, 
		    inet_ntoa(p->iph->ip_src), hostname, 0) != 0) {
			fprintf(stderr, "Failed to create trace request \n");
			return (-1);
		}
		created_trace = 1;
	} else if (strstr((char *)alert->alertmsg, "End") == NULL){
		if (create_trace_msg(&idip_trace_msg, (char *)alert->alertmsg, 
		    "192.168.11.2", hostname, 0) != 0) {
			fprintf(stderr, "Failed to create trace request \n");
			return (-1);
		}
		created_trace = 1;
	}
	/*
	 * Send the idip trace message to the firewall. The firewall	
	 * will then begin a tcpdump process and gather data
	 * from the host we have identified in this message.
	 * The firewall will notify upstream routers of the attack
	 * so they can block traffic from this host.
	 * 
	 * The use of sendto obviates the need to make an explicit
	 * connection to the receiver of this message.
	 */
	if (created_trace) {
		if (sendto(snd, (idip_message_t *)&idip_trace_msg, 
		    sizeof(idip_message_t), 0,
		    (struct sockaddr *)&snd_addr, sizeof (snd_addr)) < 0) {
			perror("could not send message to idip message layer");
		}
		created_trace = 0;
	}
	

	/*
	 * Send the do rate limiting message to the idip receiver
	 * on the firewall. This will notify the firewall to
	 * start limiting the incoming rate of packets from
	 * the hostname specified.
	 */
	if (p != NULL) {
		if (strstr((char *)alert->alertmsg, "End")) {
			if (create_do_msg(&idip_do_msg, 
			    (char *)alert->alertmsg, 
			    hostname, inet_ntoa(p->iph->ip_src), 
			    IDIP_RESTORE_RATE) != 0) { 
				fprintf(stderr, "Failed to create_do_msg \n");
				return (-1);
			}
		} else {
			/*
			 * This is a rate limiting request.
			 */
			if (create_do_msg(&idip_do_msg, 
			    (char *)alert->alertmsg, hostname,
			    inet_ntoa(p->iph->ip_src),
			    IDIP_LIMIT_RATE) != 0) { 
			    fprintf(stderr, "Failed to create_do_msg \n");
				return (-1);
			}
		} 
	} else { /* end if p != NULL */
		if (strstr((char *)alert->alertmsg, "End")) {
			if (create_do_msg(&idip_do_msg, 
			    (char *)alert->alertmsg, 
			    hostname,
			    "192.168.11.2", IDIP_RESTORE_RATE) != 0) { 
				fprintf(stderr, "Failed to create_do_msg \n");
				return (-1);
			}
		} else {
			if (create_do_msg(&idip_do_msg, 
			    (char *)alert->alertmsg, hostname,
			    "192.168.11.2", IDIP_LIMIT_RATE) != 0) { 
					fprintf(stderr, 
				        "Failed to create_do_msg \n");
					return (-1);
			}
		}
	}

	if (sendto(snd, (idip_message_t *)&idip_do_msg, 
		sizeof(idip_message_t), 0, 
		    (struct sockaddr *)&snd_addr, sizeof (snd_addr)) < 0) {
			perror("could not send message to idip message layer ");
			exit (1);
	}
if (debug) {
fprintf(stderr, "Sent IDIP do message to firewall\n");
fflush(stderr);
}
	/*
	 * Log the data we just sent to the firewall.
	 */
	
	if (p) {	
		if (p->iph && (p->tcph || p->udph || p->icmph)) {
			switch (p->iph->ip_proto) {
			case IPPROTO_TCP:
				fprintf(log, "TCP from: %s:%d ",
				    inet_ntoa(p->iph->ip_src), 
				    ntohs(p->tcph->th_sport));
				fprintf(log,"to: %s:%d\n", 
				    inet_ntoa(p->iph->ip_dst), 
				    ntohs(p->tcph->th_dport));
				break;

			case IPPROTO_UDP:
				fprintf(log,"UDP from: %s:%d ", 
				    inet_ntoa(p->iph->ip_src), 
				    ntohs (p->udph->uh_sport));
				fprintf(log,"to: %s:%d\n", 
				    inet_ntoa (p->iph->ip_dst), 
				    ntohs (p->udph->uh_dport));
				break;

			case IPPROTO_ICMP:
				fprintf (log,"ICMP type: %d code: %d from: %s ",
				    p->icmph->type, p->icmph->code, 
				    inet_ntoa (p->iph->ip_src));
				fprintf (log,"to: %s\n", 
				    inet_ntoa (p->iph->ip_dst));
				break;
			}
		}
	}
	close (snd);
	fflush(log);
	return (1);
}

void help_msg(char *prog) 
{

	fprintf(stdout,"Usage : %s [-d ] -h <hostname> -p <port no> -l  \
	    <log file name> \n", prog);
	fprintf(stdout, " option d turn on the debugging mode \n");
	fprintf(stdout, " deafult port no is %d \n", IDIP_APP_PORT);
	return;
}

int
main (int argc, char *argv[])
{
	struct sockaddr_un  bogus;
	Alertpkt	alert;
	/*
	 * Default host name for firewall.
	 */
	char 		*logfile = NULL;
	int  		portno = 0;
	char		*host = NULL;

	socklen_t 	len = sizeof (struct sockaddr_un);
	int 		sockfd;
	int 		c;
	struct 		sockaddr_un snortaddr;
 
	while ((c = getopt(argc, argv, "dhpl")) != EOF) {
		switch(c) {
		case 'd':
			debug = 1;
			fprintf(stdout, "debugging mode turned on \n");
			break;
		case 'h':
			/*
			 * Hostname of idip firewall machine.
			 */
			host = argv[optind];
			if (debug)
				fprintf(stdout, "hostname : %s \n",host);
			break;
		case 'p':
			/*
			 * Port on which idip message receiver listens
			 * on firewall for do and trace messages.
			 */
			portno = atoi(argv[optind]);
			if (debug)
				fprintf(stdout, "port # %d \n", portno);
			break;
		case 'l':
			logfile = argv[optind];
			if (debug) 
				fprintf(stdout," log file is %s \n",logfile);
			break;
		default:
			fprintf(stdout, "unrecognized arg >%s< %d \n", 
			    optarg, optind);
			help_msg(argv[0]);
			break;
		}
	}
	/* open the log file */
	if (logfile) {
		if ((logfp = fopen(logfile, "a+")) == NULL) {
			(void) fprintf(stderr, "Unable to open the log file"
			    " for logging");
			return (0);
		}
	}

	/* open the socket for checking the connections */

	unlink(UNSOCK_FILE);
  
	if ((sockfd = socket(AF_UNIX, SOCK_DGRAM, 0)) < 0) {
		perror("socket");
		exit (1);
	}

	bzero(&snortaddr, sizeof (snortaddr));
	snortaddr.sun_family = AF_UNIX;
	strcpy(snortaddr.sun_path, UNSOCK_FILE);

	if (bind (sockfd, (struct sockaddr *) &snortaddr, 
		sizeof (snortaddr)) < 0) {
		perror ("bind");
		exit (1);
	}
	    

	/* check if any alerts have been received */
	while (recvfrom(sockfd, (void *)&alert, sizeof (alert), 0, 
	    (struct sockaddr *) &bogus, &len) > 0) {
		fprintf(stderr, "Got a message to process \n");
		fprintf(stderr, "message received %s\n", alert.alertmsg);
		if (analyzePacket(&alert, logfp, host, IDIP_APP_PORT) < 0) {
			fprintf(stderr, "Unable to respond to alert \n");
			perror ("recvfrom");
			break; 
		}
	}
 
	fclose(logfp);
	close(sockfd);
	unlink (UNSOCK_FILE);
	return (0);
}
